# 
# Program: Traffic light simulation 
#          (Queue example)
# 
# The traffic light has this behavior:
# Green north/south for 40 seconds
# Pause 2 seconds
# Green east/west for 45 seconds
# Pause 2 seconds
# Repeat
# 
# The traffic behaves this way:
# A northbound car arrives at the traffic light
#   every 3 seconds;
# Southbound, every 5 seconds;
# Eastbound, every 4 seconds;
# Westbound, every 6 seconds.
# All times are approximate (random).
# Assume no cars turn at the light.
# 
# Cars pass through the light at a rate of
# one per second.
# 
# Let's run for 8900 seconds (100 full cycles or
# more than two hours) and answer these questions: 
# How long on the average is each line of cars 
# when the light turns green? What is the average 
# wait time in seconds? What is the longest wait 
# time?
# 

    
# Direction constants

NORTH, SOUTH, EAST, WEST = 0, 1, 2, 3
dirs = %w[North South East West]

# Probabilities for car arriving 
# from each direction:

p = Array.new(4)
p[NORTH] = 1.0/3.0
p[SOUTH] = 1.0/5.0
p[EAST]  = 1.0/4.0
p[WEST]  = 1.0/6.0

# Queues:

waiting = Array.new(4)
waiting[NORTH] = Queue.new
waiting[SOUTH] = Queue.new
waiting[EAST]  = Queue.new
waiting[WEST]  = Queue.new

lengths = [0, 0, 0, 0]  # How long is queue 
                        # when light turns green?
greens  = [0, 0, 0, 0]  # How many times did
                        # light turn green?
times   = [0, 0, 0, 0]  # How long did cars wait?
ncars   = [0, 0, 0, 0]  # Count cars through light.
maxtime = [0, 0, 0, 0]  # Max wait time?

# Looping...

time=0
while time < 8900

  change = true  # Light changed
  for time in time..time+40          # North/south green
    # Enqueue all arrivals
    for dir in NORTH..WEST
      waiting[dir].enqueue(time) if rand < p[dir]
    end

    # Record queue lengths, counts
    if change 
      for dir in NORTH..SOUTH
        lengths[dir] += waiting[dir].length
        greens[dir] += 1
      end
      change = false
    end
    
    # N/S can leave, one per second...
    for dir in NORTH..SOUTH
      if !waiting[dir].empty?
        car = waiting[dir].dequeue
        wait = time - car
        ncars[dir] += 1
        times[dir] += wait
        maxtime[dir] = [maxtime[dir],wait].max
      end
    end
  end

  for time in time..time+2           # Yellow/red
    # Nothing happens...
  end

  change = true  # Light changed
  for time in time..time+45          # East/west green
    # Enqueue all arrivals
    for dir in NORTH..WEST
      waiting[dir].enqueue(time) if rand < p[dir]
    end

    # Record queue lengths, counts
    if change 
      for dir in EAST..WEST
        lengths[dir] += waiting[dir].length
        greens[dir] += 1
      end
      change = false
    end
    
    # N/S can leave, one per second...
    for dir in EAST..WEST
      if !waiting[dir].empty?
        car = waiting[dir].dequeue
        wait = time - car
        ncars[dir] += 1
        times[dir] += wait
        maxtime[dir] = [maxtime[dir],wait].max
      end
    end
  end

  for time in time..time+2           # Yellow/red
    # Nothing happens...
  end

end

# Display results...

puts "Average queue lengths:"
for dir in NORTH..WEST
  printf "  %-5s %6.1f\n", dirs[dir],
         lengths[dir]/greens[dir].to_f
end

puts "Max wait times:"
for dir in NORTH..WEST
  printf "  %-5s %4d\n", dirs[dir],
         maxtime[dir]
end

puts "Average wait times:"
for dir in NORTH..WEST
  printf "  %-5s %6.1f\n", dirs[dir],
         times[dir]/ncars[dir].to_f
end
